/*  
 * Copyright (c) 2010 Ant Kutschera, maxant
 * 
 * This file is part of the maxant PersistentStateMachine.
 * 
 * maxant PersistentStateMachine is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant PersistentStateMachine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with maxant PersistentStateMachine.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.maxant.demo.scxml;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import org.apache.commons.logging.Log;
import org.apache.commons.scxml.io.SCXMLParser;
import org.apache.commons.scxml.model.ModelException;
import org.apache.commons.scxml.model.SCXML;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import uk.co.maxant.demo.scxml.util.AbstractStateMachine;
import uk.co.maxant.demo.scxml.util.LogHelper;

/**
 * this is the class that you need to implement for your own states.
 * 
 * demo of states used when purchasing a travel ticket, 
 * showing how to impl the {@link AbstractStateMachine}.
 *
 * @author   Ant Kutschera
 */
public final class StateMachineDemo extends AbstractStateMachine {

    private static final String XML_CONFIG_FILENAME = "demoStateChartConfig.xml";

	/** TRANSITION. */
    private static final String BOOK = "BOOK";

    /** TRANSITION. */
    private static final String PAYMENT = "PAYMENT";

    /** TRANSITION. */
    private static final String CANCEL = "CANCEL";

    /** TRANSITION. */
    private static final String REFUND = "REFUND";

    /** STATE. */
    public static final String RESERVED = "RESERVED";

    /** STATE. */
    public static final String BOOKED = "BOOKED";

    /** STATE. */
    public static final String PAID = "PAID";

    /** STATE. */
    public static final String CANCELLED = "CANCELLED";

    /** STATE. */
    public static final String REFUNDED = "REFUNDED";

    /** instance of model */
	private static SCXML config;

	/*
	 * read the state machine configuration, just once though, because it is quite resource intensive.
	 */
    static {
    	final Log log = LogHelper.getLog();
    	final URL configUrl = StateMachineDemo.class.getClassLoader().getResource(XML_CONFIG_FILENAME);
    	try{
    		config = SCXMLParser.parse(configUrl, new ErrorHandler(){
    		    public void error(SAXParseException exception) throws SAXException {
    		        log.error("Couldn't parse SCXML Config (" + configUrl + ")", exception);
    		    }

    		    /**
    		     * @see  ErrorHandler#fatalError(SAXParseException)
    		     */
    		    public void fatalError(SAXParseException exception) throws SAXException {
    		        log.fatal("Couldn't parse SCXML Config (" + configUrl + ")", exception);
    		    }

    		    /**
    		     * @see  ErrorHandler#warning(SAXParseException)
    		     */
    		    public void warning(SAXParseException exception) throws SAXException {
    		        log.warn("Couldn't parse SCXML Config (" + configUrl + ")", exception);
    		    }
    		});
    	}catch(ModelException e){
    		log.error("Couldn't parse SCXML Config (" + configUrl + ")", e);
    	}catch(SAXException e){
    		log.error("Couldn't parse SCXML Config (" + configUrl + ")", e);
    	}catch(IOException e){
    		log.error("Couldn't parse SCXML Config (" + configUrl + ")", e);
    	}
    }

    /**
     * creates a state machine, with states and transitions based on the given config.
     *
     * @throws  MalformedURLException
     */
    public StateMachineDemo() {
        this(null);
    }

    /**
     * creates a state machine, with states and transitions based on the given config.
     *
     * @param   idOfPersistedState  id of initial state, for cases where we are reading it from persistence.
     *
     * @throws  MalformedURLException
     */
    public StateMachineDemo(String idOfPersistedState) {
        super(config, idOfPersistedState);
    }

    /**
     * trigger the state transition.
     * 
     * @throws  ModelException
     */
    public void triggerBook() {
        trigger(BOOK);
    }

    /**
     * trigger the state transition.
     * 
     * @throws  ModelException
     */
    public void triggerPayment() {
        trigger(PAYMENT);
    }

    /**
     * trigger the state transition.
     * 
     * @throws  ModelException
     */
    public void triggerCancel() {
        trigger(CANCEL);
    }

    /**
     * trigger the state transition.
     * 
     * @throws  ModelException
     */
    public void triggerRefund() {
        trigger(REFUND);
    }

}
