/*  
 * Copyright (c) 2010 Ant Kutschera, maxant
 * 
 * This file is part of the maxant PersistentStateMachine.
 * 
 * maxant PersistentStateMachine is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant PersistentStateMachine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with maxant PersistentStateMachine.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.maxant.demo.scxml;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;

import static org.junit.Assert.*;
import org.junit.Test;

import uk.co.maxant.demo.scxml.StateMachineDemo;
import uk.co.maxant.demo.scxml.data.SomethingPersistentWithState;
import uk.co.maxant.demo.scxml.data.SomethingPersistentWithStateDao;



/**
 * test for state machine demo.
 */
public class StateMachineDemoTest {

	private static EntityManager entityManager;

	static {
        EntityManagerFactory factory = Persistence.createEntityManagerFactory("SCXMLDemo2");
        entityManager = factory.createEntityManager();
	}

    /**
     * runs a positive test.
     */
    @Test
    public void testStateMachinePositive1() {

        StateMachineDemo sm = new StateMachineDemo();

        // check initial state
        assertEquals(StateMachineDemo.RESERVED, sm.getCurrentState());

        // check state transition
        sm.triggerBook();
        assertEquals(StateMachineDemo.BOOKED, sm.getCurrentState());
    }

    /**
     * runs a negative test.
     */
    @Test
    public void testStateMachineNegative1() {

        try {
            StateMachineDemo sm = new StateMachineDemo();

            // check initial state
            assertEquals(StateMachineDemo.RESERVED, sm.getCurrentState());

            // check state transition
            sm.triggerRefund();
            fail("shouldnt be allowed, because we cannot go from reserved -> booked!");
        } catch (IllegalStateException e) {
            // excellent, this is perfect!
        } catch (Throwable t) {
            t.printStackTrace();
            fail("didnt expect this! " + t);
        }
    }

    /**
     * runs a negative test.
     */
    @Test
    public void testStateMachineNegative2() {

        try {
            StateMachineDemo sm = new StateMachineDemo(StateMachineDemo.PAID);

            // check initial state
            assertEquals(StateMachineDemo.PAID, sm.getCurrentState());
            
            //we explicitly set the state? but surely it should base whether its final on the config?
            assertTrue(sm.isCurrentStateFinal());

            //check that it actually lets it do a valid transition
            sm.triggerRefund();

            // check new state is actually set after transition
            assertEquals(StateMachineDemo.REFUNDED, sm.getCurrentState());

            // check state transition
            try {
            	sm.triggerBook();
            	fail("shouldnt be allowed, because we cannot go from refunded -> booked!");
            }catch(IllegalStateException e){
            	//perfect, thats expected here
            }
        } catch (IllegalArgumentException e) {
        	fail("didnt expect this! " + e);
        } catch (Throwable t) {
            t.printStackTrace();
            fail("didnt expect this! " + t);
        } 
    }
    
    @Test
    public void testSettingAllStates(){
    	try{
    		assertEquals(StateMachineDemo.RESERVED, new StateMachineDemo(StateMachineDemo.RESERVED).getCurrentState());
    		assertEquals(StateMachineDemo.BOOKED, new StateMachineDemo(StateMachineDemo.BOOKED).getCurrentState());
    		assertEquals(StateMachineDemo.CANCELLED, new StateMachineDemo(StateMachineDemo.CANCELLED).getCurrentState());
    		assertEquals(StateMachineDemo.PAID, new StateMachineDemo(StateMachineDemo.PAID).getCurrentState());
    		assertEquals(StateMachineDemo.REFUNDED, new StateMachineDemo(StateMachineDemo.REFUNDED).getCurrentState());
    		
    		//finally check, that if we dont provide an initial state, that it still gets set to reserved
    		assertEquals(StateMachineDemo.RESERVED, new StateMachineDemo().getCurrentState());

    	}catch(IllegalStateException e){
    		e.printStackTrace();
    		fail(e.getMessage());
    	}
    	
    	try{
    		new StateMachineDemo("non existant state");
    		fail("expected to get an exception actually, because the requested state doesnt exist!");
    	}catch(IllegalStateException e){
    		//perfect! as expected.
    	}

    }
    
    /**
     * test having multiple instances of the state machine to ensure they are thread safe.
     */
    @Test
    public void testMultipleMachines(){
		StateMachineDemo sm1 = new StateMachineDemo();
		StateMachineDemo sm2 = new StateMachineDemo();
		
		sm1.triggerBook();

		assertEquals(StateMachineDemo.BOOKED, sm1.getCurrentState());
		assertEquals(StateMachineDemo.RESERVED, sm2.getCurrentState());
		
		sm2.triggerBook();
		
		assertEquals(StateMachineDemo.BOOKED, sm1.getCurrentState());
		assertEquals(StateMachineDemo.BOOKED, sm2.getCurrentState());
		
		sm1.triggerPayment();
		
		assertEquals(StateMachineDemo.PAID, sm1.getCurrentState());
		assertEquals(StateMachineDemo.BOOKED, sm2.getCurrentState());
		
		sm2.triggerCancel();
		
		assertEquals(StateMachineDemo.PAID, sm1.getCurrentState());
		assertEquals(StateMachineDemo.CANCELLED, sm2.getCurrentState());
    	
    }
    
    @Test
    public void testStatePersistence(){
    		
    	SomethingPersistentWithStateDao dao = new SomethingPersistentWithStateDao(entityManager);
    	
    	SomethingPersistentWithState o = new SomethingPersistentWithState();
        entityManager.getTransaction().begin();
		dao.persist(o);
		entityManager.getTransaction().commit();
        entityManager.clear(); //to ensure we really get something from the DB on the next read
		
		int id = o.getIdPk();
		System.out.println("Persisted an object with ID " + id);
		
		//now load it, and change the state and persist it again
		o = dao.readById(id);

		assertEquals(StateMachineDemo.RESERVED, o.getState());
	
		o.triggerBook();
        entityManager.getTransaction().begin();
        dao.persist(o);
        entityManager.getTransaction().commit();
        entityManager.clear(); //to ensure we really get something from the DB on the next read

		//now load it again, and check that the latest state is set
		o = dao.readById(id);

		assertEquals(StateMachineDemo.BOOKED, o.getState());

		//finally, repeat state transition, this time from non-initial state
		o.triggerPayment();
        entityManager.getTransaction().begin();
        dao.persist(o);
        entityManager.getTransaction().commit();
        entityManager.clear(); //to ensure we really get something from the DB on the next read
		assertEquals(StateMachineDemo.PAID, o.getState());
		o = dao.readById(id);
		assertEquals(StateMachineDemo.PAID, o.getState());
    }

}
